@@ -1,19 +1,22 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 import datetime
+from collections import defaultdict
 
 from sql import Literal
-from sql.aggregate import Count
+from sql.aggregate import Count, Max
 
 from trytond.i18n import gettext
-from trytond.model import ModelView, ModelSQL, MatchMixin, fields, \
-    sequence_ordered
-from trytond.pyson import Eval, If, Bool
+from trytond.model import (
+    Index, MatchMixin, ModelSQL, ModelView, fields, sequence_ordered)
+from trytond.modules.currency.fields import Monetary
+from trytond.modules.product import (
+    ProductDeactivatableMixin, price_digits, round_price)
 from trytond.pool import Pool, PoolMeta
+from trytond.pyson import Bool, Eval, If, TimeDelta
+from trytond.tools import (
+    grouped_slice, is_full_text, lstrip_wildcard, reduce_ids)
 from trytond.transaction import Transaction
-from trytond.tools import lstrip_wildcard
-
-from trytond.modules.product import price_digits
 
 from .exceptions import PurchaseUOMWarning
 
@@ -26,19 +29,20 @@ class Template(metaclass=PoolMeta):
         states={
             'invisible': (~Eval('purchasable', False)
                 | ~Eval('context', {}).get('company')),
-            },
-        depends=['purchasable'])
-    purchase_uom = fields.Many2One('product.uom', 'Purchase UOM', states={
+            })
+    purchase_uom = fields.Many2One(
+        'product.uom', "Purchase UoM",
+        states={
             'invisible': ~Eval('purchasable'),
             'required': Eval('purchasable', False),
             },
-        domain=[('category', '=', Eval('default_uom_category'))],
-        depends=['purchasable', 'default_uom_category'])
+        domain=[('category', '=', Eval('default_uom_category', -1))],
+        help="The default Unit of Measure for purchases.")
 
     @fields.depends('default_uom', 'purchase_uom', 'purchasable')
     def on_change_default_uom(self):
         try:
-            super(Template, self).on_change_default_uom()
+            super().on_change_default_uom()
         except AttributeError:
             pass
         if self.default_uom:
@@ -50,27 +54,25 @@ class Template(metaclass=PoolMeta):
 
     @classmethod
     def view_attributes(cls):
-        return super(Template, cls).view_attributes() + [
+        return super().view_attributes() + [
             ('//page[@id="suppliers"]', 'states', {
                     'invisible': ~Eval('purchasable'),
                     })]
 
     def product_suppliers_used(self, **pattern):
-        # Skip rules to test pattern on all records
-        with Transaction().set_user(0):
-            template = self.__class__(self)
-        for product_supplier in template.product_suppliers:
+        for product_supplier in self.product_suppliers:
             if product_supplier.match(pattern):
                 yield product_supplier
 
     @classmethod
-    def write(cls, *args):
+    def check_modification(cls, mode, templates, values=None, external=False):
         pool = Pool()
         Warning = pool.get('res.user.warning')
-        actions = iter(args)
-        for templates, values in zip(actions, actions):
-            if not values.get("purchase_uom"):
-                continue
+
+        super().check_modification(
+            mode, templates, values=values, external=external)
+
+        if mode == 'write' and values.get("purchase_uom"):
             for template in templates:
                 if not template.purchase_uom:
                     continue
@@ -83,7 +85,6 @@ class Template(metaclass=PoolMeta):
                     if Warning.check(name):
                         raise PurchaseUOMWarning(
                             name, gettext('purchase.msg_change_purchase_uom'))
-        super(Template, cls).write(*args)
 
     @classmethod
     def copy(cls, templates, default=None):
@@ -117,31 +118,94 @@ class Product(metaclass=PoolMeta):
     product_suppliers = fields.One2Many(
         'purchase.product_supplier', 'product', "Suppliers",
         domain=[
-            ('template', '=', Eval('template')),
+            ('template', '=', Eval('template', -1)),
             ],
         states={
             'invisible': (~Eval('purchasable', False)
                 | ~Eval('context', {}).get('company')),
-            },
-        depends=['template', 'purchasable'])
+            })
     purchase_price_uom = fields.Function(fields.Numeric(
             "Purchase Price", digits=price_digits), 'get_purchase_price_uom')
+    last_purchase_price_uom = fields.Function(fields.Numeric(
+            "Last Purchase Price", digits=price_digits),
+        'get_last_purchase_price_uom')
 
     @classmethod
     def get_purchase_price_uom(cls, products, name):
         quantity = Transaction().context.get('quantity') or 0
         return cls.get_purchase_price(products, quantity=quantity)
 
+    @classmethod
+    def get_last_purchase_price_uom(cls, products, name=None):
+        pool = Pool()
+        Company = pool.get('company.company')
+        Currency = pool.get('currency.currency')
+        Date = pool.get('ir.date')
+        Line = pool.get('purchase.line')
+        Purchase = pool.get('purchase.purchase')
+        UoM = pool.get('product.uom')
+
+        transaction = Transaction()
+        context = transaction.context
+        cursor = transaction.connection.cursor()
+        purchase = Purchase.__table__()
+        line = Line.__table__()
+
+        line_ids = []
+        where = purchase.state.in_(['confirmed', 'processing', 'done'])
+        supplier = context.get('supplier')
+        if supplier:
+            where &= purchase.party == supplier
+        for sub_products in grouped_slice(products):
+            query = (line
+                .join(purchase, condition=line.purchase == purchase.id)
+                .select(
+                    Max(line.id),
+                    where=where & reduce_ids(
+                        line.product, map(int, sub_products)),
+                    group_by=[line.product]))
+            cursor.execute(*query)
+            line_ids.extend(i for i, in cursor)
+        lines = Line.browse(line_ids)
+
+        uom = None
+        if context.get('uom'):
+            uom = UoM(context['uom'])
+
+        currency = None
+        if context.get('currency'):
+            currency = Currency(context['currency'])
+        elif context.get('company'):
+            currency = Company(context['company']).currency
+        date = context.get('purchase_date') or Date.today()
+
+        prices = defaultdict(lambda: None)
+        for line in lines:
+            default_uom = line.product.default_uom
+            if not uom or default_uom.category != uom.category:
+                product_uom = default_uom
+            else:
+                product_uom = uom
+            unit_price = UoM.compute_price(
+                line.unit, line.unit_price, product_uom)
+            if currency and line.purchase.currency != currency:
+                with Transaction().set_context(date=date):
+                    unit_price = Currency.compute(
+                        line.purchase.currency, unit_price, currency,
+                        round=False)
+            prices[line.product.id] = round_price(unit_price)
+        return prices
+
     def product_suppliers_used(self, **pattern):
-        # Skip rules to test pattern on all records
-        with Transaction().set_user(0):
-            product = self.__class__(self)
-        for product_supplier in product.product_suppliers:
+        for product_supplier in self.product_suppliers:
             if product_supplier.match(pattern):
                 yield product_supplier
         pattern['product'] = None
         yield from self.template.product_suppliers_used(**pattern)
 
+    def _get_purchase_unit_price(self, quantity=0):
+        return
+
     @classmethod
     def get_purchase_price(cls, products, quantity=0):
         '''
@@ -153,13 +217,12 @@ class Product(metaclass=PoolMeta):
         '''
         pool = Pool()
         Uom = pool.get('product.uom')
-        User = pool.get('res.user')
+        Company = pool.get('company.company')
         Currency = pool.get('currency.currency')
         Date = pool.get('ir.date')
         ProductSupplier = pool.get('purchase.product_supplier')
         ProductSupplierPrice = pool.get('purchase.product_supplier.price')
 
-        today = Date.today()
         context = Transaction().context
         prices = {}
 
@@ -172,14 +235,16 @@ class Product(metaclass=PoolMeta):
         currency = None
         if context.get('currency'):
             currency = Currency(context['currency'])
+        elif context.get('company'):
+            currency = Company(context['company']).currency
+        date = context.get('purchase_date') or Date.today()
 
-        user = User(Transaction().user)
+        last_purchase_prices = cls.get_last_purchase_price_uom(products)
 
         for product in products:
-            prices[product.id] = product.get_multivalue('cost_price')
+            unit_price = product._get_purchase_unit_price(quantity=quantity)
             default_uom = product.default_uom
-            default_currency = (user.company.currency if user.company
-                else None)
+            default_currency = currency
             if not uom or default_uom.category != uom.category:
                 product_uom = default_uom
             else:
@@ -194,16 +259,22 @@ class Product(metaclass=PoolMeta):
                 pattern = ProductSupplierPrice.get_pattern()
                 for price in product_supplier.prices:
                     if price.match(quantity, product_uom, pattern):
-                        prices[product.id] = price.unit_price
-                        default_uom = product_supplier.uom
+                        unit_price = price.unit_price
+                        default_uom = product_supplier.unit
                         default_currency = product_supplier.currency
-            prices[product.id] = Uom.compute_price(
-                default_uom, prices[product.id], product_uom)
-            if currency and default_currency:
-                date = context.get('purchase_date') or today
-                with Transaction().set_context(date=date):
-                    prices[product.id] = Currency.compute(default_currency,
-                        prices[product.id], currency, round=False)
+                if unit_price is not None:
+                    unit_price = Uom.compute_price(
+                        default_uom, unit_price, product_uom)
+                    if currency and default_currency:
+                        with Transaction().set_context(date=date):
+                            unit_price = Currency.compute(
+                                default_currency, unit_price, currency,
+                                round=False)
+            if unit_price is None:
+                unit_price = last_purchase_prices[product.id]
+            else:
+                unit_price = round_price(unit_price)
+            prices[product.id] = unit_price
         return prices
 
     @classmethod
@@ -236,71 +307,79 @@ class Product(metaclass=PoolMeta):
         return new_products
 
 
-class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
-    'Product Supplier'
+class ProductSupplier(
+        sequence_ordered(), ProductDeactivatableMixin, MatchMixin,
+        ModelSQL, ModelView):
     __name__ = 'purchase.product_supplier'
     template = fields.Many2One(
         'product.template', "Product",
-        required=True, ondelete='CASCADE', select=True,
+        required=True, ondelete='CASCADE',
         domain=[
             If(Bool(Eval('product')),
                 ('products', '=', Eval('product')),
                 ()),
             ],
-        depends=['product'])
+        states={
+            'readonly': Eval('id', -1) >= 0,
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends={'company'})
     product = fields.Many2One(
-        'product.product', "Variant", select=True,
+        'product.product', "Variant",
         domain=[
             If(Bool(Eval('template')),
                 ('template', '=', Eval('template')),
                 ()),
             ],
-        depends=['template'])
-    party = fields.Many2One('party.party', 'Supplier', required=True,
-        ondelete='CASCADE', select=True)
-    name = fields.Char('Name', size=None, translate=True, select=True)
-    code = fields.Char('Code', size=None, select=True)
-    prices = fields.One2Many('purchase.product_supplier.price',
-            'product_supplier', 'Prices')
-    company = fields.Many2One('company.company', 'Company', required=True,
-        ondelete='CASCADE', select=True,
-        domain=[
-            ('id', If(Eval('context', {}).contains('company'), '=', '!='),
-                Eval('context', {}).get('company', -1)),
-            ])
-    lead_time = fields.TimeDelta('Lead Time')
+        states={
+            'readonly': Eval('id', -1) >= 0,
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends={'company'})
+    party = fields.Many2One(
+        'party.party', 'Supplier', required=True, ondelete='CASCADE',
+        states={
+            'readonly': Eval('id', -1) >= 0,
+            },
+        context={
+            'company': Eval('company', -1),
+            },
+        depends={'company'})
+    name = fields.Char("Name", translate=True)
+    code = fields.Char("Code")
+    prices = fields.One2Many(
+        'purchase.product_supplier.price', 'product_supplier', "Prices",
+        help="Add price for different criteria.\n"
+        "The last matching line is used.")
+    company = fields.Many2One(
+        'company.company', "Company",
+        required=True, ondelete='CASCADE')
+    lead_time = fields.TimeDelta(
+        "Lead Time",
+        domain=['OR',
+            ('lead_time', '=', None),
+            ('lead_time', '>=', TimeDelta()),
+            ],
+        help="The time from confirming the purchase order to receiving the "
+        "products.\n"
+        "If empty the lead time of the supplier is used.")
     currency = fields.Many2One('currency.currency', 'Currency', required=True,
         ondelete='RESTRICT')
-    uom = fields.Function(
-        fields.Many2One('product.uom', "UOM"), 'on_change_with_uom')
+    unit = fields.Function(
+        fields.Many2One('product.uom', "Unit"), 'on_change_with_unit')
 
     @classmethod
-    def __register__(cls, module_name):
-        transaction = Transaction()
-        cursor = transaction.connection.cursor()
-        table = cls.__table_handler__(module_name)
-        sql_table = cls.__table__()
-
-        # Migration from 5.0: add product/template
-        if (table.column_exist('product')
-                and not table.column_exist('template')):
-            table.column_rename('product', 'template')
-
-        super(ProductSupplier, cls).__register__(module_name)
-
-        # Migration from 3.8: change delivery_time inte timedelta lead_time
-        if table.column_exist('delivery_time'):
-            cursor.execute(*sql_table.select(
-                    sql_table.id, sql_table.delivery_time))
-            for id_, delivery_time in cursor.fetchall():
-                if delivery_time is None:
-                    continue
-                lead_time = datetime.timedelta(days=delivery_time)
-                cursor.execute(*sql_table.update(
-                        [sql_table.lead_time],
-                        [lead_time],
-                        where=sql_table.id == id_))
-            table.drop_column('delivery_time')
+    def __setup__(cls):
+        cls.code.search_unaccented = False
+        super().__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.code, Index.Similarity())),
+                })
 
     @staticmethod
     def default_company():
@@ -334,44 +413,60 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
                 self.currency, = row
 
     def get_rec_name(self, name):
-        if self.name:
-            name = self.name
-        elif self.product:
-            name = self.product.name
+        if not self.name and not self.code:
+            if self.product:
+                name = self.product.rec_name
+            else:
+                name = self.template.rec_name
         else:
-            name = self.template.name
-        if self.code:
-            name = '[' + self.code + '] ' + name
+            if self.name:
+                name = self.name
+            elif self.product:
+                name = self.product.name
+            else:
+                name = self.template.name
+            if self.code:
+                name = '[' + self.code + ']' + name
         return name
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        if clause[1].startswith('!') or clause[1].startswith('not '):
+        _, operator, operand, *extra = clause
+        if operator.startswith('!') or operator.startswith('not '):
             bool_op = 'AND'
         else:
             bool_op = 'OR'
-        code_value = clause[2]
-        if clause[1].endswith('like'):
-            code_value = lstrip_wildcard(clause[2])
+        code_value = operand
+        if operator.endswith('like') and is_full_text(operand):
+            code_value = lstrip_wildcard(operand)
         domain = [bool_op,
-            ('template',) + tuple(clause[1:]),
-            ('product',) + tuple(clause[1:]),
-            ('party',) + tuple(clause[1:]),
-            ('code', clause[1], code_value) + tuple(clause[3:]),
-            ('name',) + tuple(clause[1:]),
+            ('template.rec_name', operator, operand, *extra),
+            ('product.rec_name', operator, operand, *extra),
+            ('party.rec_name', operator, operand, *extra),
+            ('code', operator, code_value, *extra),
+            ('name', operator, operand, *extra),
             ]
         return domain
 
     @fields.depends(
         'product', '_parent_product.purchase_uom',
         'template', '_parent_template.purchase_uom')
-    def on_change_with_uom(self, name=None):
+    def on_change_with_unit(self, name=None):
         if self.product:
-            if self.product.purchase_uom:
-                return self.product.purchase_uom.id
+            return self.product.purchase_uom
         elif self.template:
-            if self.template.purchase_uom:
-                return self.template.purchase_uom.id
+            return self.template.purchase_uom
+
+    @property
+    def lead_time_used(self):
+        # Use getattr because it can be called with an unsaved instance
+        lead_time = getattr(self, 'lead_time', None)
+        party = getattr(self, 'party', None)
+        if lead_time is None and party:
+            company = getattr(self, 'company', None)
+            lead_time = party.get_multivalue(
+                'supplier_lead_time', company=company.id if company else None)
+        return lead_time
 
     def compute_supply_date(self, date=None):
         '''
@@ -380,10 +475,11 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         Date = Pool().get('ir.date')
 
         if not date:
-            date = Date.today()
-        if self.lead_time is None:
+            with Transaction().set_context(company=self.company.id):
+                date = Date.today()
+        if self.lead_time_used is None:
             return datetime.date.max
-        return date + self.lead_time
+        return date + self.lead_time_used
 
     def compute_purchase_date(self, date):
         '''
@@ -391,9 +487,10 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
         '''
         Date = Pool().get('ir.date')
 
-        if self.lead_time is None or date is None:
-            return Date.today()
-        return date - self.lead_time
+        if self.lead_time_used is None or date is None:
+            with Transaction().set_context(company=self.company.id):
+                return Date.today()
+        return date - self.lead_time_used
 
     @staticmethod
     def get_pattern():
@@ -406,7 +503,6 @@ class ProductSupplier(sequence_ordered(), ModelSQL, ModelView, MatchMixin):
 
 class ProductSupplierPrice(
         sequence_ordered(), ModelSQL, ModelView, MatchMixin):
-    'Product Supplier Price'
     __name__ = 'purchase.product_supplier.price'
     product_supplier = fields.Many2One('purchase.product_supplier',
             'Supplier', required=True, ondelete='CASCADE')
@@ -415,54 +511,44 @@ class ProductSupplierPrice(
         required=True,
         domain=[('quantity', '>=', 0)],
         help='Minimal quantity.')
-    unit_price = fields.Numeric('Unit Price', required=True,
-        digits=price_digits)
+    unit_price = Monetary(
+        "Unit Price", currency='currency', required=True, digits=price_digits)
 
-    uom = fields.Function(fields.Many2One('product.uom', 'UOM',
-        help="The unit in which the quantity is specified."),
-        'on_change_with_uom')
+    unit = fields.Function(fields.Many2One(
+            'product.uom', "Unit",
+            help="The unit in which the quantity is specified."),
+        'on_change_with_unit')
     currency = fields.Function(
         fields.Many2One('currency.currency', 'Currency'),
         'on_change_with_currency')
 
     @classmethod
-    def __register__(cls, module_name):
-        cursor = Transaction().connection.cursor()
-        table = cls.__table_handler__(module_name)
-        sql_table = cls.__table__()
-
-        fill_sequence = not table.column_exist('sequence')
-
-        super(ProductSupplierPrice, cls).__register__(module_name)
-
-        # Migration from 3.2: replace quantity by sequence for order
-        if fill_sequence:
-            cursor.execute(*sql_table.update(
-                    [sql_table.sequence], [sql_table.quantity]))
+    def __setup__(cls):
+        super().__setup__()
+        cls.__access__.add('product_supplier')
 
     @staticmethod
     def default_quantity():
         return 0.0
 
-    @fields.depends('product_supplier', '_parent_product_supplier.product')
-    def on_change_with_uom(self, name=None):
-        if self.product_supplier and self.product_supplier.uom:
-            return self.product_supplier.uom.id
+    @fields.depends('product_supplier', '_parent_product_supplier.unit')
+    def on_change_with_unit(self, name=None):
+        return self.product_supplier.unit if self.product_supplier else None
 
     @fields.depends('product_supplier', '_parent_product_supplier.currency')
     def on_change_with_currency(self, name=None):
-        if self.product_supplier and self.product_supplier.currency:
-            return self.product_supplier.currency.id
+        if self.product_supplier:
+            return self.product_supplier.currency
 
     @staticmethod
     def get_pattern():
         return {}
 
-    def match(self, quantity, uom, pattern):
+    def match(self, quantity, unit, pattern):
         pool = Pool()
         Uom = pool.get('product.uom')
         test_quantity = Uom.compute_qty(
-            self.product_supplier.uom, self.quantity, uom)
+            self.product_supplier.unit, self.quantity, unit)
         if test_quantity > abs(quantity):
             return False
-        return super(ProductSupplierPrice, self).match(pattern)
+        return super().match(pattern)
